# Generated by Django 5.0 on 2024-03-01 17:24

import django.db.models.deletion
import mptt.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flex_menu", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="menuitem",
            name="label_en",
            field=models.CharField(
                help_text="Display name on the website.",
                max_length=50,
                null=True,
                verbose_name="label",
            ),
        ),
        migrations.AddField(
            model_name="menuitem",
            name="label_fa",
            field=models.CharField(
                help_text="Display name on the website.",
                max_length=50,
                null=True,
                verbose_name="label",
            ),
        ),
        migrations.AlterField(
            model_name="menu",
            name="language",
            field=models.CharField(
                choices=[("en", "English"), ("fa", "Farsi")], default="en", max_length=5
            ),
        ),
        migrations.AlterField(
            model_name="menuitem",
            name="language",
            field=models.CharField(
                choices=[("en", "English"), ("fa", "Farsi")], default="en", max_length=5
            ),
        ),
        migrations.AlterField(
            model_name="menuitem",
            name="parent",
            field=mptt.fields.TreeForeignKey(
                blank=True,
                limit_choices_to={"level": 0},
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="children",
                to="flex_menu.menuitem",
                verbose_name="parent",
            ),
        ),
    ]



from mptt.managers import TreeManager


class MenuItemManager(TreeManager):
    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).filter(enabled=True).order_by("order")

    def roots(self, *args, **kwargs):
        return self.filter(level=0)



import logging

from django import template
from django.template.base import token_kwargs
from django.template.loader import get_template
from django.urls import NoReverseMatch, reverse

from ..enums import MenuTypes
from ..models import Menu, MenuItem
register = template.Library()
logger = logging.getLogger(__name__)


def flatten_nested_context(context, parent_key=""):
    flattened_dict = {}
    for item in context:
        if issubclass(item.__class__, template.context.BaseContext):
            nested_dict = flatten_nested_context(reversed(item.dicts), parent_key)
            flattened_dict.update(nested_dict)
        elif isinstance(item, dict):
            for key, value in item.items():
                new_key = f"{parent_key}.{key}" if parent_key else key
                flattened_dict[new_key] = value
        else:
            flattened_dict[parent_key] = item

    return flattened_dict


class MenuNode(template.Node):
    def __init__(self, nodeslist, extra_context):
        super().__init__()
        self.extra_context = extra_context
        self.nodeslist = nodeslist

    def render(self, context):
        resolved_context = {
            name: var.resolve(context) for name, var in self.extra_context.items()
        }
        base_template = get_template("flex_menu/navigation_menu.html")
        items = self.nodeslist.render(context)
        resolved_context["menu_active"] = "cursor-pointer active" in items
        context.update({**resolved_context, "menu_items": items})
        return base_template.render(flatten_nested_context(context.dicts))


@register.tag
def navigation_menu(parser, token):
    """
    TODO: Usage should be included in here
    """
    bits = token.split_contents()
    if len(bits) < 3:
        raise template.TemplateSyntaxError(
            "navigation_menu tag takes at least 2 arguments."
        )

    options = {}
    remaining_bits = bits[1:]
    while remaining_bits:
        option = remaining_bits.pop(0)
        if option == "with":
            value = token_kwargs(remaining_bits, parser, support_legacy=False)
            if not value:
                raise template.TemplateSyntaxError(
                    "With requires at least one argument."
                )
            options[option] = value

    namemap = options.get("with", {})
    nodelist = parser.parse(("endnavigation_menu",))
    parser.delete_first_token()

    return MenuNode(nodeslist=nodelist, extra_context=namemap)


class HtmxLinkNode(template.Node):
    def __init__(self, nodeslist, extra_context):
        super().__init__()
        self.nodeslist = nodeslist
        self.extra_context = extra_context

    def render(self, context):
        base_template = get_template("flex_menu/link.html")
        items = self.nodeslist.render(context)
        resolved_context = {
            name: var.resolve(context) for name, var in self.extra_context.items()
        }
        context.update({**resolved_context, "content": items})
        return base_template.render(flatten_nested_context(context.dicts))


@register.tag
def link(parser, token):
    bits = token.split_contents()
    if len(bits) < 3:
        raise template.TemplateSyntaxError(
            "navigation_menu tag takes at least 1 arguments."
        )

    options = {}
    remaining_bits = bits[1:]
    while remaining_bits:
        option = remaining_bits.pop(0)
        if option == "with":
            value = token_kwargs(remaining_bits, parser, support_legacy=False)
            if not value:
                raise template.TemplateSyntaxError(
                    "With requires at least one argument."
                )
            options[option] = value

    namemap = options.get("with", {})
    if "href" not in namemap:
        raise template.TemplateSyntaxError("href is a required argument.")

    nodelist = parser.parse(("endlink",))
    parser.delete_first_token()

    return HtmxLinkNode(nodeslist=nodelist, extra_context=namemap)


@register.inclusion_tag("flex_menu/navigation_link.html", takes_context=True)
def navigation_link(context, *, item):
    if not isinstance(item, MenuItem):
        raise ValueError(f"item must be a MenuItem instance {type(item)}.")

    # breakpoint()
    context.update(
        {
            "navigation_title": item.label,
            "icon": item.icon,
            "navigation_link": get_reverse(item.link),
            "navigation_extras": item.extras,
            "navigation_no_reload": item.no_reload,
            "navigation_is_active": context.request.resolver_match.view_name
            == item.link,
            "is_menuitem": item.parent is not None,
        }
    )
    
    return context


@register.inclusion_tag("flex_menu/menu.html", takes_context=True)
def load_navbar(context, menu_slug: str):
    if (menu_qs := Menu.objects.filter(slug=menu_slug, type=MenuTypes.NAVBAR)).exists():
        menu = menu_qs.first()
        context.update({"menu": menu, "items": menu.items.roots().get_cached_trees()})
        return context

    raise ValueError(f"Menu with slug {menu_slug} does not exist.")


@register.inclusion_tag("flex_menu/sidebar.html", takes_context=True)
def load_sidebar(context, menu_slug: str):
    if (
        menu_qs := Menu.objects.filter(slug=menu_slug, type=MenuTypes.SIDEBAR)
    ).exists():
        menu = menu_qs.first()
        context.update({"menu": menu, "items": menu.items.get_cached_trees()})
        return context

    return context.update({"menu": Menu.objects.none()})


def get_reverse(url):
    try:
        return reverse(url)
    except NoReverseMatch:
        return url



from django.conf import settings
from django.utils.safestring import mark_safe


class AppSettings:
    def __init__(self):
        self.settings = getattr(settings, "FLEX_MENU", {})
    
    @property
    def LANGUAGES(self):
        return self.settings.setdefault("LANGUAGES", [("en", "English"), ("fa", "Farsi")])
    
    @property
    def MAX_LEVEL(self):
        return self.settings.setdefault("MAX_LEVEL", 2)
    
    @property
    def MENU_ITEMS_CACHE_KEY(self):
        return self.settings.setdefault("ITEMS_CACHE_KEY", "menu_items")

    @property
    def MENU_ITEMS_CACHE_TIMEOUT(self):
        return self.settings.setdefault("ITEMS_CACHE_TIMEOUT", 60 * 60 * 24)

    @property
    def TEMPLATE_SETTINGS(self):
        TEMPLATE_SETTINGS = {
            "a_navigation_htmx": (
                'hx-swap="multi:#page-body:outerHTML,#navigation-bar:outerHTML,#jsScripts:outerHTML,#modals:outerHTML,#cssStyles:outerHTML" '
                'hx-push-url="true" hx-indicator="#page-loading" hx-boost="true"'
            ),
        }

        TEMPLATE_SETTINGS = {key: mark_safe(value) for key, value in TEMPLATE_SETTINGS.items()}
        return self.settings.setdefault("TEMPLATE_SETTINGS", TEMPLATE_SETTINGS)
    

app_settings = AppSettings()



from django.urls import get_resolver, URLPattern, URLResolver


def get_project_urls():
    def getter(urls, patterns=None, namespaces=None):
        patterns = [] if patterns is None else patterns
        namespaces = [] if namespaces is None else namespaces

        if not urls:
            return
        l = urls[0]
        if isinstance(l, URLPattern):
            yield patterns + [str(l.pattern)], namespaces + [l.name], l.callback
        elif isinstance(l, URLResolver):
            yield from getter(
                l.url_patterns, patterns + [str(l.pattern)], namespaces + [l.namespace]
            )
        yield from getter(urls[1:], patterns, namespaces)

    for pattern in getter(get_resolver().url_patterns):
        url, names, view = pattern
        names = [n for n in names if n is not None]
        if all(names):
            yield {
                "names": names,
                "view_name": ":".join(names),
                "url": "".join(url),
                "view": view,
            }
            
            
def get_url_choices():
    seen = set()
    for u in sorted(get_project_urls(), key=lambda u: u["view_name"]):
        if (name := u["view_name"]) and name not in seen:
            seen.add(name)
            yield (name, f'{name} ({u["url"]})')
            


from modeltranslation.translator import translator, TranslationOptions
from .models import MenuItem


class MenuItemTranslationOptions(TranslationOptions):
    fields = ('label',)

translator.register(MenuItem, MenuItemTranslationOptions)



import logging
import re

from django.contrib.auth.models import Permission
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from mptt.models import MPTTModel, TreeForeignKey

from .default_settings import app_settings
from .enums import MenuTypes
from .managers import MenuItemManager

logger = logging.getLogger()


class AbstractLocalized(models.Model):
    language = models.CharField(
        choices=app_settings.LANGUAGES, default="en", max_length=5
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Menu(AbstractLocalized):
    name = models.CharField(_("name"), max_length=50, unique=True)
    type = models.CharField(_("menu type"), choices=MenuTypes.choices, max_length=2)
    slug = models.SlugField(_("slug"), unique=True)
    attributes = models.JSONField(
        _("attributes"),
        default=dict,
    )
    icon = models.CharField(max_length=50, blank=True, null=True)
    enabled = models.BooleanField(_("enabled"), default=True)

    class Meta:
        verbose_name = _("Menu")
        verbose_name_plural = _("Menus")

    def __str__(self):
        return "{}".format(self.name)

    @property
    def parsed_attributes(self):
        return mark_safe("\n".join([f'{k}="{v}"' for k, v in self.attributes.items()]))


class MenuItem(MPTTModel, AbstractLocalized):
    MAX_LEVEL = app_settings.MAX_LEVEL
    label = models.CharField(
        _("label"), max_length=50, help_text=_("Display name on the website.")
    )
    slug = models.SlugField(
        _("slug"), unique=True, help_text=_("Unique identifier for the menu.")
    )
    menu = models.ForeignKey(
        Menu,
        on_delete=models.CASCADE,
        related_name="items",
        verbose_name=_("menu"),
    )
    parent = TreeForeignKey(
        "self",
        limit_choices_to={"level": 0},
        verbose_name=_("parent"),
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        related_name="children",
    )
    link = models.CharField(_("link"), max_length=255, blank=True, default="#")
    order = models.IntegerField(_("order"))
    icon = models.CharField(_("icon"), max_length=50, blank=True, null=True)
    enabled = models.BooleanField(_("enabled"), default=True)
    login_required = models.BooleanField(
        _("login required"),
        default=False,
        help_text=_(
            "If this is checked, only logged-in users will be able to view the item."
        ),
    )
    permissions = models.ManyToManyField(
        Permission,
        verbose_name=_("permissions"),
        blank=True,
        help_text=_(
            "If empty, the menu item will be publicly visible, otherwise only users with at least one of the selected permissions could see it."  # noqa
        ),
    )
    no_reload = models.BooleanField(default=True)
    extras = models.CharField(
        _("extras"),
        max_length=255,
        blank=True,
        null=True,
        help_text=mark_safe(
            _(
                'Comma separated list of extra-attributes, e.g.: <code>icon="fa fa-user",data-tooltip="Go home!"</code>'
            )
        ),
    )

    objects = MenuItemManager()
    all_objects = models.Manager()

    active = False
    with_active = False

    def __str__(self):
        return "{}".format(self.label)

    class Meta:
        verbose_name = _("MenuItem")
        verbose_name_plural = _("MenuItems")
        ordering = ["order"]

    class MPTTMeta:
        order_insertion_by = ("order",)

    def clean(self):
        super().clean()
        if (self.level or 0) > self.MAX_LEVEL - 1:
            raise ValidationError("Further nested menu is not supported.")

    @cached_property
    def has_children(self):
        return self.children.exists()

    def extras_dict(self):
        try:
            return dict(
                [re.sub("\"|'", "", i).strip() for i in s.split("=")]
                for s in self.extras.split(",")
            )
        except Exception:
            return {}



from modeltranslation.admin import TranslationAdmin

from mptt.admin import MPTTModelAdmin

from django import forms
from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from .utils import get_url_choices
from .models import Menu, MenuItem


@admin.register(Menu)
class MenuAdmin(admin.ModelAdmin):
    list_display = ["id", "name", "slug", "language", "type", "items"]
    list_editable = ["slug"]
    list_filter = ["type", "language"]

    def items(self, obj):
        return obj.items.count()


@admin.register(MenuItem)
class MenuItemAdmin(MPTTModelAdmin, TranslationAdmin):
    list_display = (
        "slug",
        "label",
        "icon",
        "order",
        "menu",
        "link",
        "parent",
        "login_required",
        "enabled",
    )
    list_filter = (
        "parent",
        "enabled",
        "login_required",
    )
    list_editable = ["label", "icon", "order", "parent", "link", "enabled"]
    search_fields = ("label",)
    filter_horizontal = ("permissions",)
    prepopulated_fields = {"slug": ("label",)}
    fieldsets = (
        (
            _("Main"),
            {
                "fields": (
                    "menu",
                    "parent",
                    "label",
                    "slug",
                    "link",
                    "icon",
                    "order",
                    "extras",
                ),
                "description": _("A menu item without parent identifies a new menu."),
            },
        ),
        (
            _("Visibility"),
            {
                "fields": (
                    "enabled",
                    "login_required",
                    "permissions",
                ),
                "classes": ("tab-fs-permissions",),
                "description": _(
                    "Yuo can decide to disable or restrict the visibility of this voice and consequently of all its children. Keep in mind that also if a child is publicly visible, but this voice requires a login, then the child will not be visible to not logged users. The same happens for permissions restrictions."
                ),
            },
        ),
    )

    def get_queryset(self, request):
        qs = self.model.all_objects.all()

        ordering = self.get_ordering(request)
        if ordering:
            qs = qs.order_by(*ordering)
        return qs



# Generated by Django 4.2.6 on 2023-10-07 13:51

from django.db import migrations, models
import django.db.models.deletion
import mptt.fields


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="Menu",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "language",
                    models.CharField(
                        choices=[("fa", "Persian"), ("en", "English")],
                        default="en",
                        max_length=5,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "name",
                    models.CharField(max_length=50, unique=True, verbose_name="name"),
                ),
                (
                    "type",
                    models.CharField(
                        choices=[("sb", "Sidebar"), ("nb", "Navbar")],
                        max_length=2,
                        verbose_name="menu type",
                    ),
                ),
                ("slug", models.SlugField(unique=True, verbose_name="slug")),
                (
                    "attributes",
                    models.JSONField(default=dict, verbose_name="attributes"),
                ),
                ("icon", models.CharField(blank=True, max_length=50, null=True)),
                ("enabled", models.BooleanField(default=True, verbose_name="enabled")),
            ],
            options={
                "verbose_name": "Menu",
                "verbose_name_plural": "Menus",
            },
        ),
        migrations.CreateModel(
            name="MenuItem",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "language",
                    models.CharField(
                        choices=[("fa", "Persian"), ("en", "English")],
                        default="en",
                        max_length=5,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "label",
                    models.CharField(
                        help_text="Display name on the website.",
                        max_length=50,
                        verbose_name="label",
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        help_text="Unique identifier for the menu.",
                        unique=True,
                        verbose_name="slug",
                    ),
                ),
                (
                    "link",
                    models.CharField(
                        blank=True, default="#", max_length=255, verbose_name="link"
                    ),
                ),
                ("order", models.IntegerField(verbose_name="order")),
                (
                    "icon",
                    models.CharField(
                        blank=True, max_length=50, null=True, verbose_name="icon"
                    ),
                ),
                ("enabled", models.BooleanField(default=True, verbose_name="enabled")),
                (
                    "login_required",
                    models.BooleanField(
                        default=False,
                        help_text="If this is checked, only logged-in users will be able to view the item.",
                        verbose_name="login required",
                    ),
                ),
                ("no_reload", models.BooleanField(default=True)),
                (
                    "extras",
                    models.CharField(
                        blank=True,
                        help_text='Comma separated list of extra-attributes, e.g.: <code>icon="fa fa-user",data-tooltip="Go home!"</code>',
                        max_length=255,
                        null=True,
                        verbose_name="extras",
                    ),
                ),
                ("lft", models.PositiveIntegerField(editable=False)),
                ("rght", models.PositiveIntegerField(editable=False)),
                ("tree_id", models.PositiveIntegerField(db_index=True, editable=False)),
                ("level", models.PositiveIntegerField(editable=False)),
                (
                    "menu",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="items",
                        to="flex_menu.menu",
                        verbose_name="menu",
                    ),
                ),
                (
                    "parent",
                    mptt.fields.TreeForeignKey(
                        blank=True,
                        limit_choices_to={"parent": None},
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="children",
                        to="flex_menu.menuitem",
                        verbose_name="parent",
                    ),
                ),
                (
                    "permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="If empty, the menu item will be publicly visible, otherwise only users with at least one of the selected permissions could see it.",
                        to="auth.permission",
                        verbose_name="permissions",
                    ),
                ),
            ],
            options={
                "verbose_name": "MenuItem",
                "verbose_name_plural": "MenuItems",
                "ordering": ["order"],
            },
        ),
    ]
